<?php
namespace DaReactions\Pages;
use DaReactions\Options;
use Closure;
/**
 * Class SettingsPage
 * @package DaReactions\Pages
 */
class SettingsPage
{
    /**
     * @var string $current_tab
     * Current tab identifier
     *
     * @since 3.1.1
     */
    protected $current_tab = '';
    /**
     * @var array $navigation
     * List of tabs
     */
    protected $navigation;
    /**
     * @var Options $options
     * The Options instance to manage page settings
     *
     * @since 1.0.0
     */
    protected $options;
    /**
     * @var string $options_group
     * The name of the group for saved options
     *
     * @since 1.0.0
     */
    protected $options_group;
    /**
     * @var string $options_page
     * The slug of the settings page in wich the options are managed
     *
     * @since 1.0.0
     */
    protected $options_page;
    /**
     * Page constructor.
     *
     * @param string $options_group
     * @param string $section
     *
     * @since 1.0.0
     */
    public function __construct( string $options_group, string $section)
    {
        $this->options_group = $options_group . '_' . $section;
        $this->options_page = $options_group . '_' . $section . '_page';
        $this->options = Options::getInstance($section);
    }
    /**
     * Renders the main form for this page
     *
     * @since 1.0.0
     */
    public function displayPage(): void {
        ?>
        <form action="<?= 'options.php' ?>" method='post'>
            <?php
            settings_fields($this->options_group);
            $this->displayNavigationTabs();
            do_settings_sections($this->options_page);
            submit_button();
            ?>
        </form>
        <?php
        $this->displayLoaderOverlay();
    }
    public function displayLoaderOverlay(): void {
        ?>
        <div class="da-spinner-wrapper">
            <div class="da-spinner"></div>
        </div><?php
    }
    public function displayNavigationTabs(): void {
        echo $this->getNavigationTabs();
    }
    /**
     * Render function for navigation tabs
     *
     * @since 3.1.1
     */
    public function getNavigationTabs(): string {
        if (!isset($this->navigation) || count($this->navigation) < 2) {
            return '';
        }
        $page = $this->options_group . '_settings';
        $op = '<h2 class="nav-tab-wrapper">';
        foreach ($this->navigation as $key => $nav) {
            $url = admin_url("admin.php?page=$page&tab=$key");
            $active_class = $this->current_tab === $key ? ' nav-tab-active' : '';
            $op .= '<a href="' . $url . '" class="nav-tab' . $active_class . '">' . $nav['title'] . '</a >';
        }
        $op .= '</h2>';
        return $op;
    }
    /**
     * Return generic checkbox render function
     *
     * @param $field_id
     * @param $message
     * @param array $attributes
     *
     * @return Closure
     */
    public function makeCheckboxRenderer($field_id, $message, array $attributes = array()): callable {
        return function () use ($field_id, $message, $attributes) {
            $field_name = $this->options->getFieldName($field_id);
            $saved_value = $this->options->getOption($field_id);
            $checked = isset($saved_value) && $saved_value === 'on';
            $additional_attributes = '';
            foreach ($attributes as $key => $value) {
                $additional_attributes .= ' ' . $key . '="' . $value . '"';
            }
            ?>
            <p>
                <input
                        type="hidden"
                        name="<?= $field_name ?>"
                        value="off"/>
                <input id="id_<?= $field_name ?>" type="checkbox"
                       name="<?= $field_name ?>" <?php checked($checked, 1) ?>
                    <?= $additional_attributes ?>
                       value="on"/>
                <label for="id_<?= $field_name ?>"><?= $message ?></label>
            </p>
            <?php
        };
    }
	/**
	 * Return generic section render function
	 *
	 * @param $message
	 *
	 * @return Closure
	 *
	 * @since 3.1.1
	 */
    public function makeSectionRenderer($message): callable {
        return static function () use ($message) {
            echo "<p>$message</p>";
            echo '<hr>';
        };
    }
    /**
     * Return generic select field render function
     *
     * @param $field_id
     * @param $default
     * @param $options
     * @param array $attributes
     *
     * @return Closure
     */
    public function makeSelectRenderer($field_id, $default, $options, array $attributes = array()): callable {
        return function () use ($field_id, $default, $options, $attributes) {
            $field_name = $this->options->getFieldName($field_id);
            $saved_value = $this->options->getOption($field_id, $default);
            $additional_attributes = '';
            foreach ($attributes as $value => $label) {
                $additional_attributes .= ' ' . $value . '="' . $label . '"';
            }
            ?>
            <p>
            <select
                    name="<?= $field_name ?>"
                <?= $additional_attributes ?>>
                <?php foreach ($options as $value => $label) { ?>
                    <option
                            value="<?= $value ?>" <?= ($saved_value === $value ? 'selected' : '') ?>>
                        <?= $label ?>
                    </option>
                <?php } ?>
            </select>
            <?php
        };
    }
    /**
     * Return generic input field render function
     *
     * @param $field_id
     * @param $default
     * @param array $attributes
     *
     * @return Closure
     */
    public function makeTextfieldRenderer($field_id, $default, array $attributes = array()): callable {
        return function () use ($field_id, $default, $attributes) {
            $field_name = $this->options->getFieldName($field_id);
            $saved_value = $this->options->getOption($field_id, $default);
            $additional_attributes = '';
            foreach ($attributes as $key => $value) {
                $additional_attributes .= ' ' . $key . '="' . $value . '"';
            }
            ?>
            <p>
                <input
                        id="id_<?= $field_name ?>"
                    <?= $additional_attributes ?>
                        name="<?= $field_name ?>"
                        value="<?= $saved_value ?>"/>
            </p>
            <?php
        };
    }
}
